#ifndef DKQ_H
#define DKQ_H

#include "alias.h"
struct matrix;
struct vector;
struct ivector;

/**
   class dkq defines plate quadrilateral finite element based
   on the Kirchhoff plate theory for thin plates
   
   JK, 20. 10. 2019
*/

class dkq
{
 public:
  dkq (void);
  ~dkq (void);
  
  void auxdata (vector &x,vector &y,vector &l,vector &sx,vector &sy,vector &nx,vector &ny);
  double approx (double xi,double eta,vector &nodval);

  void geom_matrix (matrix &gm,double xi, double eta,vector &x,vector &y,vector &sx,vector &sy,vector &nx,vector &ny,vector &l,double &jac);
  void stiffness_matrix (long eid,long ri,long ci,matrix &sm,vector &x,vector &y);
  void res_stiffness_matrix (long eid,matrix &sm);
  
  
  ///  number of DOFs on the element
  long ndofe;
  ///  number of nodes on one element
  long nne;
  ///  total number of components of stress and strain tensors
  long tncomp;
  ///  number of components for graphic purposes
  long gncomp;
  ///  total number of integration points on element
  long tnip;
  ///  array containing numbers of components of stress and strain tensors
  long *ncomp;
  ///  array containing cumulative numbers of components of stress and strain tensors
  long *cncomp;
  ///  number of approximated functions on the element
  long napfun;
  ///  number of edges
  long ned;
  ///  number of nodes on one edge
  long nned;
  ///  array containing orders of numerical integration of stiffness matrix
  long **intordsm;
  ///  order of integration of mass matrix
  long intordmm;
  ///  order of integration on edges
  long intordb;
  ///  array of numbers of integration points in sets
  long **nip;
  ///  number of blocks
  long nb;
  ///  stress/strain state
  strastrestate ssst;

};

#endif
