#ifndef LAYSHELL_H
#define LAYSHELL_H

#include "alias.h"
#include "iotools.h"
#include "strretalg.h"
#include "intpoints.h"
#include "matrix.h"
struct matrix;
struct vector;
struct atsel;

/**
   Class layshell defines layered model for shells.
   
   Josef Fiedler, 9/2015
*/

class layshell
{
 public:
  layshell (void);
  ~layshell (void);
  void read (XFILE *in);
  
  void   nondiag (matrix &d, long ipp);
  void   matstiff (matrix &d, long ipp,long ido);
  void   planestiff (matrix &d, long ipp);
  void   platestiff (matrix &d, long ipp);
  void   ccnstiff (matrix &c, matrix &cn, matrix &dn, long ipp);
  
  void   nlstresses (long ipp, long im, long ido);
  void   compute_internal_forces (vector &sig, long ipp, long im, long ido, double *eps, double *k);
  
  long   compeqother (long ipp);
  long   compother (long ipp);
  
  void   backup (long ipp, double *eps, double *k, long j);
  void   restore_values (long ipp, double *eps, double *k, long j);
  void   updateval (long ipp, long im, long ido);
  double dstep_red (long ipp, long im, long ido);

  /// number of layers
  long nl;
  /// number of materials on current layer
  long *nm;
  /// array of material types
  mattype **tm;
  /// array of number of appropriate material type
  long **idm;
   /// number of iteration steps for cross-section eqilibrium
  long nli;
  /// error for iteration of cross-section eqilibrium
 // double chyba;
  /// backup of informations stored on integration point
  intpoints bcup;
};

#endif
