#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "trfelinit.h"
#include "globalt.h"
#include "globmatt.h"
#include "intpointst.h"
#include "gtopology.h"
#include "transprint.h"
#include "iotools.h"
#include "xfile.h"
#include "stacktrace.h"
#include "libtrace.h"

/**
   trfel initialization
   function reads data from input file and initiates all
   necessary variables and arrays
   
   @param argv - optinal arguments
   @param stochd - pointer to %stochdrivert
   
*/
void trfel_init (int argc, const char *argv[], stochdrivert *stochd)
{
  long i;
  XFILE *in;

  set_prgname(argv[0]);
  fprintf (stderr,"\n\n ******************************************************\n");

  Tp = new probdesct;
  Gtt = new gtopology;
  Tt = new transtop;
  Tm = new transmat;

  Tc = new transcrsec;
  Tb = new transbclc;
  Outdt = new outdrivert;
  Kmat = NULL;
  Cmat = NULL;
  Jmat = NULL;
  Bmat = NULL;

  if (process_argst(argc, argv))
  {
    delete Tp;
    delete Gtt;
    delete Tt;
    delete Tm;
    delete Tc;
    delete Tb;
    delete Outdt;
    abort();
  }

  in = xfopen(argv[1],"r");
  if (in==NULL){
    fprintf (stderr,"\n Input file has not been specified.");
    fprintf (stderr,"\n Try it again!\n\n");
    abort ();
  }
  in->warning    = 1;
  in->ignorecase = 1;
  if ((Tp->kwdsw == nokwd) || (Tp->kwdsw == kwd_outd))
    in->kwdmode = ignore;
  else
    in->kwdmode = sequent_mode;
  //in->kwdmode = ignore;
  //in->kwdmode = sequent_mode;

  Outt = fopen ("trfel.log","wt");
  
  //  basic problem data
  Tp->read (in);
  
  if (Tp->kwdsw == kwd_full)
    in->kwdmode = sequent_mode;
  else
    in->kwdmode = ignore;
  
  //  input of node and element data
  Tt->read (in);
  
  // mesh control
  /*
  if (Tp->tprob == growing_np_problem || Tp->tprob == growing_np_problem_nonlin){
    if (Mesprt==1){
      merr=Tt->mesh_check();
      if (merr > 0){
        fprintf (stdout,"Mesh error = %ld\n\n\n",merr);
        exit(0);
      }
    }
  }
  */
  
  if (Tp->tprob == growing_np_problem || Tp->tprob == growing_np_problem_nonlin){
    Gtt->read_gf (in);
    Gtt->alloc_growstr();
  }
  
  //  input of material characteristics
  Tm->read (in);
  
  //  detection of the radiation materials
  radiation_init ();
  
  
  if (Tp->tprob == discont_nonstat_problem || Tp->tprob == discont_nonlin_nonstat_problem || Gtt->edtype==mater){
    Gtt->auto_subdom_detection (Outt);
    Tt->alloc_enodes ();
    Tt->alloc_edges ();
    Tt->edge_init_edval ();

    Tt->view_factors (Outt);
    Tt->mod_view_factors (Outt);
  }
  
  // *************************************
  //  generation of ordering of unknowns
  //  (generation of code numbers)
  //  number of degrees of freedom
  // *************************************
  Ndoft = Gtt->codenum_generation (Outt);
  if (Mesprt==1)
    fprintf (stdout,"\n the number of degrees of freedom    %ld",Ndoft);

  //  fprintf(stdout, "\n Element renumbering");
  //  Tt->sort_elements();
  //  integration point allocation and initiation
  //Tm->intpointalloc ();
  //Tm->intpointinit ();
  Tm->assemble_dof_nameord(Tp->dofname, Tp->ntm, Tp->var_dofid, Tp->tnkv);

  //  input of cross section characteristics
  Tc->read (in);
  
  //  input of boundary conditions
  Tb->read (in);

  //  prescribed values on elements investigation
  //Tt->elemprescvalue();
  //  searching for elements with nodes where source is defined
  Tb->elemsource ();
  
  //  global vectors allocations
  Lsrst = new lhsrhst;
  Lsrst->alloc ();
  //  initial conditions
  Lsrst->initcond (in);
  
  //  advective terms
  if (Tp->advect==1){
    Tm->advection_velocities (in); //05/06/2017 commented by TKr, because of missing function
    //  circumscribed balls are generated and the radii are stored
    Gtt->construct_circumscribed_balls ();
  }
  
  if (Tp->stochasticcalc){
    stochd->read (in);
    Stt = stochd;
  };
  
  if (Tp->tprob == discont_nonstat_problem || Tp->tprob == discont_nonlin_nonstat_problem){

    Tt->enodes_init ();
    Tt->edge_init ();

    //  definition of material types and id on nodes
    //Tt->node_materials ();    
  }
  
  if ((Tp->kwdsw == nokwd) || (Tp->kwdsw == kwd_pd))
    in->kwdmode = ignore;
  else
    in->kwdmode = sequent_mode;

  // input of output description
  Outdt->read(in);

  if (Tp->kwdsw == kwd_full)
    in->kwdmode = sequent_mode;
  else
    in->kwdmode = ignore;
  
  fclose (Outt);
  
  char fname[1020];
  if (*Outdt->outfn)
    filename_decomposition(Outdt->outfn,Tp->path,Tp->filename,Tp->suffix);
  else
    filename_decomposition(argv[1],Tp->path,Tp->filename,Tp->suffix);

  sprintf(fname, "%s%s.log", Tp->path, Tp->filename);
  Outt = fopen(fname,"wt");
  
  //  nodes allocation
  Tt->alloc_nodes ();
  
  
  //  some solvers require additional data
  //  if the Schur complement method or FETI method
  //  are used, the variable Ndofm is rewritten
  //  Ndofm is not changed in all other cases
  Tp->ssle->prepare_data (&Ndoft,Gtt,Outt);

  
  //if (Tp->tprob == discont_nonstat_problem){
  //Tt->node_materials ();
  //}

  //  initiation of initial values
  //added by TKr 19/9/2012
  if (Tm->initval==NULL)
    Tm->initval = new double [Tm->tnip];
    
    for (i=0;i<Tm->tnip;i++){
      Tm->initval[i]=0.0;
    }    
  
  xfclose (in);
}

void print_helpt(FILE *out, const char *prgname)
{
  fflush(out);
  fprintf(out, "\nInput file has not been specified\n");
  fprintf(out, " use : %s infile [-kwd=i]\n", prgname);
  fprintf(out, " where :\n *   infile is input file name\n");
  fprintf(out, " *   -kwd={0,1,2,3,4} is optional switch which controls keyword usage in the input file\n");
  fprintf(out, "           0 means no keywords (default value)\n");
  fprintf(out, "           1 means keywords only in the problem description section\n");
  fprintf(out, "           2 means keywords only in the outdriver section\n");
  fprintf(out, "           3 means keywords in the problem description and outdriver sections\n");
  fprintf(out, "           4 means full usage of keywords\n\n");
  abort();
}


long process_argst(int argc, const char *argv[])
{
  const char *pkwd;
  
  switch (argc)
  {
    case 0:
      fflush(stdout);
      fprintf(stderr, "\n\nError - too few arguments are used\n\n");
      print_helpt(stderr, "ptrfel|trfel");
      return 1;
    case 1:
      fflush(stdout);
      fprintf(stderr, "\n\nError - too few arguments are used\n\n");
      print_helpt(stderr, argv[0]);
      return 1;
    case 2:
      Tp->kwdsw = nokwd;
      break;
    case 3:
      // processing of optional keyword switch -kwd={0,1,2,3,4}
      pkwd = strstrcis(argv[2], "-kwd=");
      if (pkwd)
      {
        Tp->kwdsw = pkwd_sw(-1);
        sscanf(pkwd+5, "%d", (int *)&Tp->kwdsw);
        if ((Tp->kwdsw < 0) || (Tp->kwdsw > 4))
              {
          fflush(stdout);
          fprintf(stderr, "\n\nError - unknown -kwd value %d is obtained\n\n", Tp->kwdsw);
          print_helpt(stderr, argv[0]);
          return 1;
        }
      }
      else
      {
        fflush(stdout);
        fprintf(stderr, "\n\nError - unknown switch is used\n\n");
        print_helpt(stderr, argv[0]);
        return 1;
      }
      break;
    default:
      fflush(stdout);
      fprintf(stderr, "\n\nError - too many arguments is used\n\n");
      print_helpt(stderr, argv[0]);
      return 1;
  }
  return 0;
}

/**
   function prepares auxiliary data for edge detection used for radiation problems
   function assigns value 1 to the variables auxinf on gelements
   
   JK, 3.6.2011
*/
void radiation_init ()
{
  long i,j,k,nmt;
  
  //  the number of material types used on a single element
  nmt=Tp->ntm*Tp->ntm;
  
  k=0;
  //  loop over the elements
  for (i=0;i<Tt->ne;i++){
    for (j=0;j<nmt;j++){
      if (Tt->elements[i].tm[j]==radiationmater){
	k=1;
	Gtt->edtype=mater;
	break;
      }
    }
    if (k==1){
      break;
    }
  }
  
  //  loop over the elements
  for (i=0;i<Tt->ne;i++){
    for (j=0;j<nmt;j++){
      if (Tt->elements[i].tm[j]==radiationmater){
	Gtt->gelements[i].auxinf=251;
	break;
      }
    }
  }
  
}

